import textwrap

__version__ = "0.29"
__author__ = "Humphreys, Matthew P. and Delaigue, Louise and Ourradi, Yasmina"


def hello():
    greeting = textwrap.dedent(
        r"""
        k  Miscellaneous tools for
        o  marine carbonate chemistry
        o  and other such things
        l  
        s  Version {}
        t  doi:10.5281/zenodo.3999292
        o  
        f  https://seaco2.group
        """.format(
            __version__
        )
    )
    print(greeting)



import PyCO2SYS as pyco2


def pH_of_analysis(
    alkalinity_sample,
    pCO2_lab=500,
    pk_HCOOH=3.43,
    pk_BPhB=3.86,
    ratio_reagent=827,
    ratio_sample=381,
    stock_BPhB_vol=50,
    stock_HCOOH_vol=50,
    **pyco2_kwargs,
):
    """Calculate the pH of the sample-reagent mixture and the fractional abundance of
    the BPhB^2- ion detected at 590 nm during QuAAtro alkalinity analysis with
    bromophenol blue (BPhB) plus formic acid (HCOOH) reagent.

    Parameters
    ----------
    alkalinity_sample : float
        Total alkalinity of the sample in μmol/kg-sample.
    pCO2_lab : float, optional
        Partial pressure of CO2 in the laboratory in μatm. The default is 500.
    pk_HCOOH : float, optional
        pK* of HCOOH. The default is 3.43 [SMP99].
    pk_BPhB : float, optional
        pK* of BPhB^- (second dissociation constant). The default is 3.86 [SMP99].
    ratio_reagent : float, optional
        Relative amount of reagent in the sample-reagent mixture. The default is 827.
    ratio_sample : float, optional
        Relative amount of sample in the sample-reagent mixture. The default is 381.
    stock_BPhB_vol : float, optional
        Volume of 0.5 g/l BPhB per litre of reagent in ml. The default is 50.
    stock_HCOOH_vol : float, optional
        Volume of 0.1 M HCOOH per litre of reagent in ml. The default is 50.
    **pyco2_kwargs
        Additional keyword arguments to pass to PyCO2SYS.

    Returns
    -------
    pH_free : float
        pH of the sample-reagent mixture on the Free scale.
    percent_detected : float
        Percentage of total BPhB in the 2- form measured at 590 nm.
    """

    # Stock solution composition --- HCOOH
    stock_HCOOH_conc = 0.1  # mol/l

    # Stock solution composition --- BPhB
    stock_BPhB_conc_mg = 500  # mg/l
    rmm_BPhB = 669.96  # g/mol
    stock_BPhB_conc = 1e-3 * stock_BPhB_conc_mg / rmm_BPhB  # mol/l

    # Reagent composition
    reagent_vol = 1000  # ml
    reagent_HCOOH = stock_HCOOH_conc * stock_HCOOH_vol / reagent_vol  # mol/l
    reagent_BPhB = stock_BPhB_conc * stock_BPhB_vol / reagent_vol  # mol/l

    # Alkalinity of reagent
    alkalinity_BPhB = -2  # mol/mol --- like H2SO4
    alkalinity_HCOOH = -1  # mol/mol --- like HF
    alkalinity_vol_reagent = (
        reagent_HCOOH * alkalinity_HCOOH + reagent_BPhB * alkalinity_BPhB
    )  # mol/l

    # Alkalinity of sample
    density_sample = 1.03  # kg/l
    alkalinity_vol_sample = 1e-6 * alkalinity_sample * density_sample  # mol/l

    # Alkalinity of sample-reagent mixture
    alkalinity_vol_mixture = (
        (alkalinity_vol_sample * ratio_sample + alkalinity_vol_reagent * ratio_reagent)
        / (ratio_sample + ratio_reagent)
        * 1e6
    )  # umol/l
    density_mixture = 1.02  # kg/l
    alkalinity_mixture = alkalinity_vol_mixture / density_mixture  # umol/kg

    # HCOOH in sample-reagent mixture
    HCOOH_vol_mixture = (
        1e6 * reagent_HCOOH * ratio_reagent / (ratio_sample + ratio_reagent)
    )  # umol/l
    HCOOH_mixture = HCOOH_vol_mixture / density_mixture  # umol/kg

    # BPhB in sample-reagent mixture
    BPhB_vol_mixture = (
        1e6 * reagent_BPhB * ratio_reagent / (ratio_sample + ratio_reagent)
    )  # umol/l
    BPhB_mixture = BPhB_vol_mixture / density_mixture  # umol/kg

    # Dissociation constants for HCOOH and BPhB
    # pk_HCOOH = 3.745  # Wikipedia
    # pk_BPhB = 3.513  # Nand & Ellwood (2018) doi:10.1002/lom3.10253
    k_HCOOH = 10**-pk_HCOOH
    k_BPhB = 10**-pk_BPhB

    # Dilute other total concentrations
    results_undiluted = pyco2.sys(
        par1=alkalinity_mixture,
        par2=pCO2_lab,
        par1_type=1,
        par2_type=4,
        total_alpha=HCOOH_mixture,
        k_alpha=k_HCOOH,
        total_beta=BPhB_mixture,
        k_beta=k_BPhB,
        **pyco2_kwargs,
    )
    totals_diluted = {
        t: results_undiluted[t] * ratio_sample / (ratio_sample + ratio_reagent)
        for t in [
            "total_sulfate",
            "total_fluoride",
            "total_borate",
            "total_silicate",
            "total_phosphate",
            "total_ammonia",
            "total_sulfide",
        ]
    }
    pyco2_kwargs = {k: v for k, v in pyco2_kwargs.items() if k not in totals_diluted}

    # Solve the sample-reagent mixture
    res = pyco2.sys(
        par1=alkalinity_mixture,
        par2=400,  # assume CO2 degassing to equilibrium with lab air
        par1_type=1,
        par2_type=4,
        total_alpha=HCOOH_mixture,
        k_alpha=k_HCOOH,
        total_beta=BPhB_mixture,
        k_beta=k_BPhB,
        **pyco2_kwargs,
        **totals_diluted,
    )

    # Return results
    pH_free = res["pH_free"]
    percent_detected = 100 * res["beta"] / res["total_beta"]
    return pH_free, percent_detected



import numpy as np


def aou_GG92(oxygen=None, temperature=25, salinity=35):
    """AOU and saturated O2 from temperature and salinity in umol/kg-sw.
    Based on Garcia & Gordon (1992), 'combined-fit' constants.

    With the corrected equation from
    Garcia, H.E., and L.I. Gordon, 1993: Erratum: Oxygen solubility in  seawater:
    better fitting equations. Limnology and Oceanography, 38(3), 656.
    """
    A0 = 5.80818
    A1 = 3.20684
    A2 = 4.11890
    A3 = 4.93845
    A4 = 1.01567
    A5 = 1.41575
    B0 = -0.00701211
    B1 = -0.00725958
    B2 = -0.00793334
    B3 = -0.00554491
    C0 = -0.000000132412
    # Scaled temperature
    ts = np.log((298.15 - temperature) / (273.15 + temperature))
    # Calculate saturated oxygen <o2s>
    o2_saturation = np.exp(
        A0
        + A1 * ts
        + A2 * ts**2
        + A3 * ts**3
        + A4 * ts**4
        + A5 * ts**5
        + salinity * (B0 + B1 * ts + B2 * ts**2 + B3 * ts**3)
        + C0 * salinity**2
    )
    # Calculate AOU
    if oxygen is not None:
        aou = o2_saturation - oxygen
        aou_percent = 100 * aou / o2_saturation
    else:
        aou = aou_percent = np.nan
    return aou, aou_percent, o2_saturation


def pH_tris_DD98(temperature=25, salinity=35):
    """pH of tris buffer following DelValls and Dickson (1998) eq. 18."""
    T = temperature + 273.15
    S = salinity
    pH = (
        (11911.08 - 18.2499 * S - 0.039336 * S**2) / T
        - 366.27059
        + 0.53993607 * S
        + 0.00016329 * S**2
        + (64.52243 - 0.084041 * S) * np.log(T)
        - 0.11149858 * T
    )
    return pH



import string
import numpy as np

lcletter = dict(zip(range(1, 27), string.ascii_lowercase))


def sigfig(x, sf):
    """Return `x` to `sf` significant figures."""
    factor = 10.0 ** np.ceil(np.log10(np.abs(x)))
    return factor * np.around(x / factor, decimals=sf)



import numpy as np
from scipy import sparse


def als(y, lam, p, niter=10):
    """Baseline correction."""
    # https://stackoverflow.com/questions/29156532/python-baseline-correction-library
    L = len(y)
    D = sparse.diags([1, -2, 1], [0, -1, -2], shape=(L, L - 2))
    D = lam * D.dot(D.transpose())
    w = np.ones(L)
    W = sparse.spdiags(w, 0, L, L)
    for _ in range(niter):
        W.setdiag(w)
        Z = W + D
        z = sparse.linalg.spsolve(Z, w * y)
        w = p * (y > z) + (1.0 - p) * (y < z)
    return z



import pandas as pd


def read_coulometer(filename):
    """Import data from Bob's coulometer.

    Parameters
    ----------
    filename : str
        The name (and path) of the file to import.

    Returns
    -------
    pd.DataFrame
        The imported file with samples identified by "number".
    """
    # Import text file and rename columns
    dic = pd.read_csv(filename).rename(
        columns={
            "Time": "datetime",
            "    Cnts": "counts",
            "    Cell_%T": "transmission",
            " Cell_mA": "current",
            " Cell_Temp": "temperature",
            " Cell_ON": "state",
            " Titrate": "titrate",
        }
    )
    dic["datetime"] = pd.to_datetime(dic.datetime)
    # Identify samples
    dic["number"] = (dic.counts < dic.counts.shift()).cumsum()
    return dic


def _get_samples(group):
    row = {}
    row["datetime_start"] = group.datetime.iloc[0]
    row["datetime_end"] = group.datetime.iloc[-1]
    row["counts"] = group.counts.iloc[-1]
    if row["counts"] != group.counts.max():
        print("WARNING! Final counts is not the greatest for the sample.")
    row["temperature_mean"] = group.temperature.mean()
    return pd.Series(row)


def get_samples(dic):
    """Get table of individual samples from imported coulometer file.

    Parameters
    ----------
    dic : pd.DataFrame
        Coulometer dataset imported with ``get_coulometer``.

    Returns
    -------
    pd.DataFrame
        Table with one row per sample.

    """
    return dic.groupby("number").apply(_get_samples)



"""
koolstof.vindta
===============

Import, process and calibrate the DIC data files produced by the Marianda VINDTA 3C.

Documentation is available online at https://koolstof.readthedocs.io/en/latest/vindta/

Code examples assume that the import convention has been followed:

  >>> import koolstof.vindta as ksv

The following subfunctions are then available.

Import data
-----------

    read_dbs
    read_logfile

Process and calibrate
---------------------

    blank_correction
    calibrate_dic
    poison_correction

Data visualisation
------------------

    plot_increments
    plot_blanks
    plot_session_blanks
    plot_dic_offset
    plot_k_dic
"""

from .read import read_dbs, read_logfile
from .get import blank_correction, calibrate_dic
from .plot import (
    plot_blanks,
    plot_dic_offset,
    plot_increments,
    plot_k_dic,
    plot_session_blanks,
)
from .process import poison_correction
from . import plot, process



import numpy as np
import pandas as pd
from matplotlib import dates as mdates


mapper_dbs = {
    "run type": "run_type",
    "i.s. temp.": "temperature_insitu",
    "sample mass": "mass_sample",
    "rep#": "rep",
    "CT": "dic",
    "factor CT": "dic_factor",
    "CV (µmol)": "cv_micromol",
    "CV (%)": "cv_percent",
    "last CRM CT": "lastcrm_dic_measured",
    "cert. CRM CT": "lastcrm_dic_certified",
    "CRM batch": "lastcrm_batch",
    "calc. mode": "mode_calculation",
    "integ. mode": "mode_integration",
    "Lat.": "latitude",
    "Long.": "longitude",
    "area#1": "area_1",
    "area#2": "area_2",
    "area#3": "area_3",
    "area#4": "area_4",
}


def read_dbs(filepath_or_buffer, encoding="unicode_escape", na_values="none", **kwargs):
    """Import the dbs file generated by a Marianda AIRICA as a pandas DataFrame.
    Any kwargs are passed to pandas.read_table.
    """
    dbs = pd.read_table(
        filepath_or_buffer, encoding=encoding, na_values=na_values, **kwargs
    )
    dbs.rename(mapper=mapper_dbs, axis=1, inplace=True)
    dbs.drop(columns="Unnamed: 32", inplace=True)
    dbs["datetime"] = pd.to_datetime(
        dbs.apply(lambda x: " ".join((x.date, x.time)), axis=1)
    )
    dbs["datenum"] = mdates.date2num(dbs.datetime)
    return dbs


mapper_LI7000 = {
    "Time": "datetime",
    "CO2B um/m": "x_CO2",
    "H2OB mm/m": "x_H2O",
    "T C": "temperature",
    "P kPa": "pressure",
    "RH %": "humidity_relative",
    "Flow V": "flow_voltage",
}


def get_licor_resolution(licor):
    """Determine the time resolution of data reported by a LI-COR, after it
    has been converted into a pandas DataFrame, and correct the 'datenum' and
    'datetime' column values in the DataFrame with second fractions.
    Input licor is generated with read_LI7000().
    """
    # Find resolution
    ldu = licor.datenum.unique()[1:-1]
    lda = licor[np.isin(licor.datenum, ldu)].datenum
    resolution = int(len(lda) / len(ldu))  # in Hz
    timestep = 1 / resolution
    # Apply correction
    start_iloc = np.where(licor.datenum == ldu[0])[0][0]
    second_fractions = np.linspace(0, 1 - timestep, resolution)
    second_fractions_start = second_fractions[-start_iloc:]
    second_fractions_middle = np.tile(second_fractions, len(ldu))
    tail_iloc = np.where(licor.datenum == ldu[-1])[0][-1] + 1
    tail_length = len(licor) - tail_iloc
    second_fractions_end = second_fractions[:tail_length]
    second_fractions = np.concatenate(
        (second_fractions_start, second_fractions_middle, second_fractions_end)
    )
    licor["datenum"] = licor.datenum + second_fractions / (60 * 60 * 24)
    licor["datetime"] = mdates.num2date(licor.datenum)
    return licor


def read_LI7000(filepath_or_buffer, skiprows=2, **kwargs):
    """Import the text files recorded by a LI-COR LI-7000 as a pandas DataFrame.
    Any kwargs are passed to pandas.read_table.
    """
    licor = pd.read_table(filepath_or_buffer, skiprows=skiprows, **kwargs)
    licor.rename(mapper=mapper_LI7000, axis=1, inplace=True)
    licor["datetime"] = pd.to_datetime(licor.datetime)
    licor["datenum"] = mdates.date2num(licor.datetime)
    return get_licor_resolution(licor)


def get_licor_samples(licor, dbs):
    """Identify different samples in a LI-COR dataset based on the dbs file,
    and remove data from before the first match.
    Assumes that both datasets are ordered with datetime ascending.
    """
    licor["dbs_ix"] = np.nan
    for dbs_ix in dbs.index:
        licor.loc[licor.datenum >= dbs.loc[dbs_ix].datenum, "dbs_ix"] = dbs_ix
    licor = licor[~np.isnan(licor.dbs_ix)]
    return licor



import pandas as pd
import numpy as np


def read_cary_oxygen(filename, us_date_format=False):
    """Import oxygen data from Agilent Cary spectrophotometer report.

    Parameters
    ----------
    filename : str
        The file name (and path) for the .csv report.
    us_date_format : bool, optional
        Whether the dates in the file are in US format (month first), by default False.

    Returns
    -------
    pd.DataFrame
        The data from the file parsed into a pandas DataFrame.
    """
    with open(filename, "r") as f:
        lines = f.read().splitlines()
    data_start = []
    data_end = []
    for i, line in enumerate(lines):
        if line == "Sample\tF\tMean\tSD\t%RSD\tReadings":
            data_start.append(i)
        elif line == "Results Flags Legend":
            data_end.append(i)
    data = []
    for start, end in zip(data_start, data_end):
        data_ = pd.read_csv(
            filename,
            skiprows=start,
            skipfooter=len(lines) - end,
            sep="\t",
            engine="python",
        )
        if us_date_format:
            mm, dd, yyyy = lines[start - 1].split()[2].split("/")
        else:
            dd, mm, yyyy = lines[start - 1].split()[2].split("/")
        HH, MM, SS = lines[start - 1].split()[3].split(":")
        try:
            if lines[start - 1].split()[4] == "PM" and HH != "12":
                HH = int(HH) + 12
        except IndexError:
            pass
        data_["datetime"] = np.datetime64(
            "{:04.0f}-{:02.0f}-{:02.0f}T{:02.0f}:{:02.0f}:{:02.0f}".format(
                int(yyyy), int(mm), int(dd), int(HH), int(MM), int(SS)
            )
        )
        data.append(data_)
    data = (
        pd.concat(data)
        .rename(
            columns={
                "Sample": "sample_name",
                "F": "flag",
                "Mean": "absorbance_mean",
                "SD": "absorbance_std",
                "%RSD": "absorbance_rsd_pct",
                "Readings": "absorbance_raw",
            }
        )
        .replace(" ", None)
        .reset_index(drop=True)
    )
    data["sample_name"] = data.sample_name.ffill()
    return data



import numpy as np
import koolstof as ks


def test_aou_GG92():
    """Does ks.aou_GG92() return the check value from GG92 Table 1?"""
    o2_saturation = ks.aou_GG92(temperature=10, salinity=35)[2]
    assert np.round(o2_saturation, decimals=3) == 274.646
    # GG92 actually give 274.647...


# test_aou_GG92()



import numpy as np
from matplotlib import pyplot as plt
from scipy import interpolate
from sklearn import cluster


def get_cluster_profile(x, y, data=None, cluster_bandwidth=5, linspace_num=100):
    """Get clusters from input data and interpolate them."""
    # Extract data from data, if required
    if data is not None:
        x = data[x]
        y = data[y]
    # Eliminate NaNs from input data
    l = ~np.isnan(x) & ~np.isnan(y)
    dx = x[l]
    dy = y[l]
    # Do MeanShift clustering of the x-variable
    ms = cluster.MeanShift(bandwidth=cluster_bandwidth)
    ms.fit(np.vstack(dx))
    # Get y-variable values at x-variable cluster centres
    cn = ms.labels_.ravel()
    cx = ms.cluster_centers_.ravel()
    cy = np.full_like(cx, np.nan)
    for i in range(len(cy)):
        cy[i] = np.mean(dy[cn == i])
    # Sort clusters
    ci = np.argsort(cx)
    cx = cx[ci]
    cy = cy[ci]
    # Generate PCHIP interpolation of clustered data
    px = np.linspace(np.min(cx), np.max(cx), num=linspace_num)
    py = interpolate.pchip(cx, cy)(px)
    return px, py, cx, cy, cn


def cluster_profile(
    x,
    y,
    data=None,
    ax=None,
    cluster_bandwidth=5,
    linspace_num=100,
    invert_xy=False,
    plot_kwargs=None,
    scatter_kwargs=None,
):
    """Get clusters from input data, interpolate them and plot the results."""
    # Extract data from data, if required
    if data is not None:
        x = data[x]
        y = data[y]
    p_xy = get_cluster_profile(
        x,
        y,
        cluster_bandwidth=cluster_bandwidth,
        linspace_num=linspace_num,
    )[:2]
    if ax is None:
        _, ax = plt.subplots()
    if plot_kwargs is None:
        plot_kwargs = {}
    if scatter_kwargs is None:
        scatter_kwargs = {}
    xy = (x, y)
    if invert_xy:
        p_xy = p_xy[::-1]
        xy = xy[::-1]
    ax_plot = ax.plot(*p_xy, **plot_kwargs)
    ax_scatter = ax.scatter(*xy, **scatter_kwargs)
    return ax, ax_plot, ax_scatter



import numpy as np


def get_data_index(dbs_row, data):
    """Get index of data in dbs to transfer metadata across for processing."""
    if dbs_row.bottle in data.station_bottleid.values:
        data_iloc = np.where(dbs_row.bottle == data.station_bottleid.values)[0]
        assert (
            len(data_iloc) == 1
        ), "Found more than one bottle ID match for {}!".format(dbs_row.bottle)
        data_index = data.index[data_iloc[0]]
    else:
        data_index = np.nan
    return data_index


def dbs2data(dbs, data, fields):
    """Transfer results from dbs into data, taking means for duplicates."""
    if isinstance(fields, str):
        fields = [fields]
    for field in fields:
        data[field] = np.nan
        data[field + "_std"] = np.nan
        data[field + "_count"] = 0
    for row in data.index:
        dr = data.loc[row]
        if dr.station_bottleid in dbs.bottle.values:
            for field in fields:
                dbs_iloc = np.where(
                    (dr.station_bottleid == dbs.bottle.values) & dbs[field + "_good"]
                )[0]
                if np.size(dbs_iloc) > 0:
                    fdata = dbs.iloc[dbs_iloc][field]
                    data.loc[row, field] = np.mean(fdata)
                    if np.size(dbs_iloc) > 1:
                        data.loc[row, field + "_std"] = np.std(fdata)
                    data.loc[row, field + "_count"] = np.size(dbs_iloc)
    return data


def poison_correction(var, sample_volume, poison_volume):
    """Apply dilution correction to var for added poison."""
    return var * (1 + poison_volume / sample_volume)



import koolstof as ks, pandas as pd

# Import files
spreadsheet_path = "tests/data/test_optode/spreadsheet.xlsx"
text_files_folder_path = "tests/data/test_optode"


def test_pH_optode():
    """Does ks.pH_optode() return a Pandas DataFrame?"""
    database = ks.pH_optode(spreadsheet_path, text_files_folder_path)
    assert isinstance(database, pd.DataFrame)


df = ks.pH_optode(spreadsheet_path, text_files_folder_path)


def test_optode_assertions():
    assert isinstance(df, pd.DataFrame)
    assert hasattr(df, "pH_NBS")
    assert hasattr(df, "filename")


# test_pH_optode()
# test_optode_assertions()



import numpy as np
from koolstof import quaatro as ksq


def test_pH_of_analysis():
    """Does the pH_of_analysis function return the correct types of results?
    Values are not checked.
    """
    # Scalar arg
    pH, pct = ksq.pH_of_analysis(2300)
    assert isinstance(pH, float)
    assert isinstance(pct, float)
    # Array arg
    pH_arr, pct_arr = ksq.pH_of_analysis(np.array([2000, 2500, 3000]))
    assert isinstance(pH_arr, np.ndarray)
    assert isinstance(pct_arr, np.ndarray)
    assert pH > np.min(pH_arr)
    assert pH < np.max(pH_arr)
    assert pct > np.min(pct_arr)
    assert pct < np.max(pct_arr)
    # Scalar arg with kwargs
    pH_kws, pct_kws = ksq.pH_of_analysis(
        2300,
        pk_HCOOH=3.43,
        pk_BPhB=3.86,
        ratio_reagent=827,
        ratio_sample=381,
        stock_BPhB_vol=50,
        stock_HCOOH_vol=50,
        temperature=20,
        salinity=30,
        total_phosphate=10,
        total_silicate=50,
    )
    assert isinstance(pH_kws, float)
    assert isinstance(pct_kws, float)


# test_pH_of_analysis()



"""Import and parse the data files produced by the Marianda AIRICA."""

import pandas as pd
import numpy as np
from matplotlib import dates as mdates
from scipy import stats
import seaborn as sns
from calkulate.density import seawater_1atm_MP81
import matplotlib.pyplot as plt


# extract data from dbs
mapper_dbs = {
    "run type": "run_type",
    "i.s. temp.": "temperature_insitu",
    "sample mass": "mass_sample",
    "rep#": "rep",
    "CT": "dic",
    "factor CT": "dic_factor",
    "CV (µmol)": "cv_micromol",
    "CV (%)": "cv_percent",
    "last CRM CT": "lastcrm_dic_measured",
    "cert. CRM CT": "lastcrm_dic_certified",
    "CRM batch": "lastcrm_batch",
    "calc. mode": "mode_calculation",
    "integ. mode": "mode_integration",
    "Lat.": "latitude",
    "Long.": "longitude",
    "area#1": "area_1",
    "area#2": "area_2",
    "area#3": "area_3",
    "area#4": "area_4",
}


def read_dbs(filepath_or_buffer, encoding="unicode_escape", na_values="none", **kwargs):
    """Import the dbs file generated by a Marianda AIRICA as a pandas DataFrame.
    All kwargs are passed to pandas.read_table.
    """
    dbs = pd.read_table(
        filepath_or_buffer, encoding=encoding, na_values=na_values, **kwargs
    )
    dbs.rename(mapper=mapper_dbs, axis=1, inplace=True)
    if "Unnamed: 32" in dbs:
        dbs.drop(columns="Unnamed: 32", inplace=True)
    dbs["datetime"] = pd.to_datetime(
        dbs.apply(lambda x: " ".join((x.date, x.time)), axis=1)
    )
    dbs["datenum"] = mdates.date2num(dbs.datetime)
    dbs["volume_sample"] = (dbs.mass_sample / dbs.density).round()  # microlitres
    return dbs


def process_airica(
    crm_val,
    db,
    dbs_filepath,
    results_file_path_and_name=None,
    draw_figure=True,
):
    """Process AIRICA raw data by extracting data from .dbs file and
    adding it to .xlsx file, calculating conversion factor from CRMs and
    computing TCO2 values.
    """

    # import ".dbs" file
    data = read_dbs(dbs_filepath)

    # add ".dbs" data to ".xlsx"
    db = pd.merge(left=db, right=data, how="left", left_on="name", right_on="bottle")

    # check that ".dbs" bottle = ".xlsx" name and drop "bottle" column
    if db["name"].equals(db["bottle"]):
        print("SUCCESSFUL DBS IMPORT")
        db = db.drop(columns=["bottle"])
    else:
        KeyError
        print("ERROR: mismatch between dbs and xlsx files")

    # recalculate density
    db["density_analysis"] = np.nan
    db["density_analysis"] = seawater_1atm_MP81(
        temperature=db.temperature, salinity=db.salinity_rws
    )

    # average areas with all areas and only last 3 areas
    db["area_av_4"] = (db.area_1 + db.area_2 + db.area_3 + db.area_4) / 4
    db["area_av_3"] = (db.area_2 + db.area_3 + db.area_4) / 3

    # calculate DIC * density * sample_v
    db["CT_d_sample_v"] = crm_val * db.density_analysis * db.sample_v

    # create columns to hold conversion factor (CF) values
    db["a_3"] = np.nan
    db["a_4"] = np.nan
    db["b_3"] = np.nan
    db["b_4"] = np.nan

    # calc CRM coeff factor
    def get_CF(db):
        """Calculate conversion factor CF for each analysis batch."""
        L = db.location == "CRM"
        b_3, a_3 = stats.linregress(db.area_av_3[L], db.CT_d_sample_v[L])[:2]
        b_4, a_4 = stats.linregress(db.area_av_3[L], db.CT_d_sample_v[L])[:2]
        return pd.Series({"a_3": a_3, "a_4": a_4, "b_3": b_3, "b_4": b_4})

    db_cf = db.groupby(by=["analysis_batch"]).apply(get_CF)

    # assign CRM a and b to samples based on analysis batch
    db["a_3"] = db_cf.loc[db.analysis_batch.values, "a_3"].values
    db["a_4"] = db_cf.loc[db.analysis_batch.values, "a_4"].values
    db["b_3"] = db_cf.loc[db.analysis_batch.values, "b_3"].values
    db["b_4"] = db_cf.loc[db.analysis_batch.values, "b_4"].values

    # calculate TCO2 values
    db["TCO2_3"] = np.nan
    db["TCO2_4"] = np.nan
    db["TCO2_3"] = ((db.b_3 * db.area_av_3) + db.a_3) / (
        db.density_analysis * db.sample_v
    )
    db["TCO2_4"] = ((db.b_4 * db.area_av_4) + db.a_4) / (
        db.density_analysis * db.sample_v
    )

    if draw_figure:
        # plot regression
        f, ax = plt.subplots(figsize=(8, 6.5), dpi=300)
        sns.set_style("darkgrid")
        sns.set_context("paper", font_scale=2)
        sns.set(font="Verdana", font_scale=1)
        sns.despine(f, left=True, bottom=True)

        L = db.location == "CRM"
        sns.regplot(
            x=db.area_av_3[L],
            y=db.CT_d_sample_v[L],
            ci=False,
            color="xkcd:primary blue",
        )

        # add R2 to graph
        r2 = stats.linregress(db.area_av_3[L], db.CT_d_sample_v[L])[2]
        r2s = str(round(r2, 2))
        text = "$R^2$ = " + r2s
        ax.text(
            30000,
            4500000,
            text,
            horizontalalignment="left",
            verticalalignment="center",
            fontsize=15,
        )
        plt.tight_layout()

    # save results as text file
    if results_file_path_and_name is not None:
        db.to_csv(results_file_path_and_name, index=None)

    return db



"""Miscellaneous tools for marine carbonate chemistry and other such things."""

from . import (
    airica,
    coulometer,
    crm,
    infrared,
    meta,
    misc,
    molar,
    optode,
    parameterisations,
    plot,
    quaatro,
    spectro,
    vindta,
)
from .meta import __author__, __version__, hello
from .misc import lcletter, sigfig
from .vindta import (
    read_dbs,
    read_logfile,
    plot_session_blanks,
    plot_blanks,
    plot_k_dic,
    plot_dic_offset,
    poison_correction,
)
from .optode import pH_optode
from .plot import get_cluster_profile, cluster_profile
from .parameterisations import aou_GG92, pH_tris_DD98
from .spectro.oxygen import read_cary_oxygen



"""Import and parse the data files produced by the PyroScience pH optode."""

import pandas as pd
import os
import numpy as np
from scipy import stats


def pH_optode(spreadsheet_path, text_files_folder_path, cutoff=1200, end_secs=120):
    """Import the txt files generated by Pyro Workbench as a pandas dataframe.
    Process data to get mean pH with a slope closest to 0. Generate table
    with results and stats. Optional kwargs include an option to choose a
    cutoff point in seconds to keep the last X seconds of data (default = 1200 sec),
    and an option to compute pH for the last few seconds of measurements (
    default = 120 sec)."""

    # Import spreadsheet with filenames
    db = pd.read_excel(spreadsheet_path, skiprows=[1])

    # Create list of text files
    file_list = [
        file
        for file in os.listdir(text_files_folder_path)
        if file in db.filename.values
    ]

    # Create loop to extract data
    data = {}
    for file in file_list:
        path = text_files_folder_path + "/{}/{}.txt"
        fname = path.format(file, file)
        data[file] = pd.read_table(fname, skiprows=22, encoding="unicode_escape")

    # Clean up datasets
    # Rename headers of df inside dict
    rn = {
        "Date [A Ch.1 Main]": "date",
        "Time [A Ch.1 Main]": "time",
        " dt (s) [A Ch.1 Main]": "sec",
        "pH [A Ch.1 Main]": "pH",
        "Sample Temp. (°C) [A Ch.1 CompT]": "temp",
    }
    for file in file_list:
        data[file].rename(columns=rn, inplace=True)

        # Drop useless (empty) columns
        data[file].drop(
            columns=[
                "Unnamed: 23",
                "Unnamed: 24",
                "Unnamed: 25",
                "Unnamed: 26",
                "Unnamed: 27",
                "Unnamed: 28",
                "Unnamed: 29",
            ],
            inplace=True,
        )

        # Remove last rows including nans
        L = data[file].pH.isnull()
        data[file] = data[file][~L]

        # Only keep the last X min of data
        if data[file].sec.max() > cutoff:
            L = data[file].sec > (data[file].sec.max() - cutoff)
            data[file] = data[file][L]

    # Make a copy of the data
    data_c = data.copy()

    # Create a table to hold the results
    results = pd.DataFrame({"filename": file_list})
    results["date"] = np.nan
    results["time"] = np.nan
    results["pH_NBS_raw"] = np.nan
    results["pH_NBS_raw_median"] = np.nan
    results["pH_end"] = np.nan
    results["pH_end_median"] = np.nan
    results["slope"] = np.nan
    results["lowest_ix"] = np.nan
    results["pH_NBS"] = np.nan
    results["pH_NBS_median"] = np.nan
    results["pH_NBS_stderr"] = np.nan
    results["pH_NBS_std"] = np.nan
    results["pH_NBS_intercept"] = np.nan

    # Compute the mean and median of the last 2min of measurements
    # using all datapoints inside file
    for file in file_list:
        L = data[file].sec > data[file].sec.max() - end_secs
        F = results.filename == file
        results.loc[F, "pH_end"] = data[file][L].pH.mean()
        results.loc[F, "pH_end_median"] = data[file][L].pH.median()

        # Do a linear regression on all data inside sample file
        slope = stats.linregress(data_c[file].sec, data_c[file].pH)[0]

        # Create column to hold slope results
        data[file]["slope_here"] = np.nan

        # Loop through the data to find the slope closest to 0
        for i in data[file].index[:-5]:
            print("processing...")
            slope = stats.linregress(data_c[file].sec, data_c[file].pH)[0]
            data_c[file] = data_c[file].drop(data_c[file].index[0])
            data_c[file].sort_values(by="sec")
            data[file].loc[i, "slope_here"] = np.abs(slope)
            print(file)

        # Find index of lowest absolute slope
        lowest_ix = data[file].slope_here.idxmin()

        # Store time for pH of slope closest to 0 in df
        results.loc[F, "time"] = data[file].time.loc[lowest_ix][:5]

        # Store temperature for pH of slope closest to 0
        results.loc[F, "temperature"] = data[file].temp.loc[lowest_ix]

        # Store the mean(raw pH) in df
        results.loc[F, "pH_NBS_raw"] = data[file].pH.mean()

        # Store the median (raw pH) in df
        results.loc[F, "pH_NBS_raw_median"] = data[file].pH.median()

        # Store the index of the lowest absolute slope in df
        results.loc[F, "lowest_ix"] = lowest_ix

        # Store the slope from lowest_ix in df
        results.loc[F, "slope"] = data[file].slope_here.loc[lowest_ix]

        # Store the mean (pH slope 0) in df
        results.loc[F, "pH_NBS"] = data[file].pH.loc[lowest_ix:].mean()

        # Store the median (pH slope 0) in df
        results.loc[F, "pH_NBS_median"] = data[file].pH.loc[lowest_ix:].median()

        # Store the standard error for pH slope 0 in df
        results.loc[F, "pH_NBS_stderr"] = stats.linregress(
            data[file].sec.loc[lowest_ix:], data[file].pH.loc[lowest_ix:]
        )[4]

        # Store the standard deviation for pH slope 0 in df
        results.loc[F, "pH_NBS_std"] = (data[file].pH.loc[lowest_ix:]).std()

        # Store the intercept for pH slope 0 in df
        results.loc[F, "pH_NBS_intercept"] = stats.linregress(
            data[file].sec.loc[lowest_ix:], data[file].pH.loc[lowest_ix:]
        )[1]

        # Add extra information
        results.loc[F, "date"] = db.date

    return results



import pandas as pd, numpy as np
from koolstof import vindta as ksv

logfile_fname = "tests/data/logfile_20200407.bak"
dbs_fname = "tests/data/2018_Aug_RWS_CO2.dbs"
logfile = ksv.read_logfile(logfile_fname, methods=["3C standard", "3C standardRWS"])
dbs = ksv.read_dbs(dbs_fname)
sessions = ksv.blank_correction(dbs, logfile)
dbs["dic_certified"] = np.where(dbs.bottle.str.startswith("CRM"), 2029.19, np.nan)
ksv.calibrate_dic(dbs, sessions)


def test_read_dbs():
    """Is the dbs correctly imported and are columns dropped (or not)?"""
    dbs = ksv.read_dbs(dbs_fname)
    assert isinstance(dbs, pd.DataFrame)
    assert "dic_cell_id" in dbs
    assert "run_type" not in dbs
    dbs = ksv.read_dbs(dbs_fname, drop_cols=False)
    assert isinstance(dbs, pd.DataFrame)
    assert "dic_cell_id" in dbs
    assert "run_type" in dbs


def test_get_logfile_index():
    dbs = ksv.read_dbs(dbs_fname)
    assert "logfile_index" not in dbs
    ksv.get.get_logfile_index(dbs, logfile)
    assert "logfile_index" in dbs


def test_get_sample_blanks():
    dbs = ksv.read_dbs(dbs_fname)
    ksv.get.get_sample_blanks(dbs, logfile)
    assert "blank_here" in dbs


def test_get_session_blanks():
    dbs = ksv.read_dbs(dbs_fname)
    sessions = ksv.get.get_session_blanks(dbs, logfile)
    assert isinstance(sessions, pd.DataFrame)
    assert "blank_here" in dbs


def test_get_counts_corrected():
    dbs = ksv.read_dbs(dbs_fname)
    ksv.get.get_counts_corrected(dbs, logfile)
    assert "counts_corrected" in dbs
    dbs = ksv.read_dbs(dbs_fname)
    ksv.get.get_sample_blanks(dbs, logfile)
    ksv.get.get_counts_corrected(dbs)
    assert "counts_corrected" in dbs


def test_blank_correction():
    dbs = ksv.read_dbs(dbs_fname)
    sessions = ksv.blank_correction(dbs, logfile)
    assert isinstance(sessions, pd.DataFrame)
    assert "counts_corrected" in dbs


def test_get_standard_calibrations():
    dbs = ksv.read_dbs(dbs_fname)
    ksv.blank_correction(dbs, logfile)
    dbs["dic_certified"] = np.nan
    dbs.loc[dbs.bottle.str.startswith("CRM"), "dic_certified"] = 2029.19
    ksv.get.get_standard_calibrations(dbs)
    assert "k_dic_here" in dbs


def test_calibrate_dic():
    dbs = ksv.read_dbs(dbs_fname)
    sessions = ksv.blank_correction(dbs, logfile)
    dbs["dic_certified"] = np.nan
    dbs.loc[dbs.bottle.str.startswith("CRM"), "dic_certified"] = 2029.19
    ksv.calibrate_dic(dbs, sessions)
    assert "k_dic_mean" in sessions
    assert "k_dic" in dbs
    assert "dic" in dbs
    assert ~dbs.dic.isnull().any()


def test_plots():
    ksv.plot_increments(dbs, logfile)
    ksv.plot_session_blanks(dbs, sessions, sessions.index[0])
    ksv.plot_blanks(dbs, sessions)
    ksv.plot_k_dic(dbs, sessions)
    ksv.plot_dic_offset(dbs, sessions)
    dbs.loc[
        (dbs.dic_cell_id == "C_Aug14-18_0708")
        & ((dbs.bottle == "CRM#171") | (dbs.bottle == "CRM#171B")),
        "k_dic_good",
    ] = False
    ksv.calibrate_dic(dbs, sessions)
    ksv.plot_k_dic(dbs, sessions, show_ignored=False)
    ksv.plot_dic_offset(dbs, sessions)


# test_read_dbs()
# test_get_logfile_index()
# test_get_sample_blanks()
# test_get_session_blanks()
# test_get_counts_corrected()
# test_blank_correction()
# test_get_standard_calibrations()
# test_calibrate_dic()
# test_plots()
